#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ascii.h"
#include "../tools/molecular.h"
#include "../tools/uvspec.h"
#include "../src_f/f77-uscore.h"

static int closest_above(float lambda, float* lambda_raw, int n_crs); 
static int closest_below(float lambda, float* lambda_raw, int n_crs);

/*      SUBROUTINE crs_o3(crs, error, first, lambda, n_lambda, ntemp, 
 *     $     path, temper, type)  */
/*
*
*BOD
*DESCRIPTION
* Calculates ozone cross sections as a function of
* wavelength and temperature.
*USAGE
**FORTRAN
*     CALL crs_o3( crs, error, lambda, n_lambda, ntemp, path,
*    $             temper, type)  
**PERL
*INPUT
**first
*     logical
*      If true, temperature coefficients for ozone will be 
*      calculated at wavelengths lambda. If first is true, it will
*      be returned as false, just in case you should forget.
*      If false, the temperature coefficients from earlier call will be
*      used. 
*      first should be true the first time crs_o3 is called. It should 
*      also be true if lambda is changed. It does not need to be true
*      if temper is changed.
**lambda
*     real array(*)
*     Wavelengths (nm) at which the ozone cross section is to be calculated.
**n_lambda
*     integer
*     Number of wavelengths lambda.
**ntemp
*     integer
*     Number of different temperatures (levels in the atmosphere) 
**path
*     character string
*     The location of the Radtran data files, e.g. 
*     /usr/local/lib/Radtran/data
**temper
*     real array(*)
*     The temperatures (levels in the atmosphere) at which
*     the ozone cross section is to be calculated. In Kelvin.
**type
*     integer
*     The ozone cross section data to be used:
***1     Molina and Molina (1986)
***2     Daumont et al. (1992), Malicet et al. (1995)
***3     Paur and Bass (1985)
*OUTPUT
**crs
*     real array(temper,lambda)
*     The ozone cross section as a function of temper and lambda.
**error
*     integer
*     Error flag. Error checking performed only if first is true.
***0     Everything ok.
***1     Error occurred during reading of data file.
***2     n_lambda == 0 and ntemp != 0
***3     n_lambda != 0 and ntemp == 0
*ERROR_MESSAGES
*ROUTINES_CALLED
*FILES
*     crs_o3_mol_cf.dat           Reference 1 data file
*     crs_o3_dau_cf.dat           Reference 2 data file
*     crs_o3_pab_cf.dat           Reference 3 data file
*     consts.inc               Include file
*     param.inc                Include file
*REFERENCES
*     (1) Molina and Molina, 'Absolute Absorption Cross Sections of
*         Ozone in the 185- to 350-nm wavelength Range', JGR, vol 91,
*         pp. 14,501-14,508, 1986.
*
*     (2) Daumont D, Brion J, Charbonnier J and Malicet J, 
*         "Absorption Cross-Section at Room Temperature",
*         J.Atmos.Chem. no. 15, vol. 145 (1992)
*
*     (3) Paur and Bass, (1985) 'The ultraviolet cross sections of ozone, II, 
*         results and temperature dependence', in: Atmospheric Ozone, proceedings 
*         of the Quadrennial Ozone Symposium, Halkidiki, Greece. 
*         C. Zerefos and A. Ghazi (eds), pp.611-616, D. Reidel, Hingham, MA 
*
*EOD
*
*
*     References:
*     
*     WMO (1986), 'Atmospheric Ozone 1985', WMO Report No. 16.
*
*     */
int crs_o3 (crs_out_struct *crs_out, int first, float *lambda, 
	   int n_lambda, int ntemp, char *path, float *temper,
	   int type)
{
  int ia=0, ib=0, it=0, iv=0;
  int n_crs=0, status=0, max_columns=0, min_columns=0, rows=0;
  static int n_coeff=0;
  static float fact=0;
  char filenm[FILENAME_MAX+200]="", fullnm[FILENAME_MAX+200]="";
  static float *lambda_raw=NULL, **coeff_crs=NULL;
  float **data=NULL;
  float a=0.0, b=0.0, gradient=0.0;

  if ( first ) {
    if (n_lambda <= 0 && ntemp != 0)
      return 2;
    else { 
      if (n_lambda != 0 &&  ntemp <= 0)
	return 3;
    }

    switch (type) {
    case O3_MOLINA:
      strcpy(filenm,"crs/crs_o3_mol_cf.dat");
      n_coeff = 3;
      fact     = 1.E-20;
      break;
    case O3_DAUMONT:
      strcpy(filenm,"crs/crs_o3_dau_cf.dat");
      n_coeff = 3;
      fact     = 1.E-20;
      break;
    case O3_BASSPAUR:
      strcpy(filenm,"crs/crs_o3_pab_cf.dat");
      n_coeff = 3;
      fact     = 1.E-20;
      break;
    }
    
    strcpy(fullnm,path);
    strcat(fullnm,filenm);

    status = ASCII_file2float (fullnm, &rows, &max_columns, 
			       &min_columns, &data); 
    if (status!=0) { 
      fprintf (stderr, " ... error %d opening file %s\n", status, fullnm); 
      return status; 
    } 

    n_crs  = rows;
    
    lambda_raw = (float *) calloc ((size_t) rows, sizeof (float));
    lambda_raw = ASCII_column_float (data, rows, 0);
    status  = ASCII_calloc_float(&coeff_crs, rows, max_columns);
    if (status!=0) { 
      fprintf (stderr, " ... error %d allocating memory\n", status); 
      return status; 
    } 

    coeff_crs[0] =  ASCII_column_float (data, rows, 1);
    coeff_crs[1] =  ASCII_column_float (data, rows, 2);
    coeff_crs[2] =  ASCII_column_float (data, rows, 3);

    /* free memory */
    ASCII_free_float(data, rows);

  } /* if (first) */
  
  for (iv=0; iv<n_lambda; iv++) {
    ia=closest_above(lambda[iv], lambda_raw, n_crs);
    ib=closest_below(lambda[iv], lambda_raw, n_crs);

    if (ia > (n_crs-1) || ib<0) {
      fprintf(stderr,"crs_o3, cannot handle this case yet\n");
      fprintf(stderr,"lambda: %f n_crs: %d ia: %d ib: %d\n",lambda[iv], n_crs, ia, ib);
      return 9;
    }

    for (it=0; it<ntemp; it++) {
      a = coeff_crs[0][ia] + 
	coeff_crs[1][ia] * (temper[it]-273.15) + 
	coeff_crs[2][ia]*(temper[it]-273.15)*(temper[it]-273.15);
      if (ia == ib) {
	crs_out->crs[it][O3][iv] = a*fact;
      }
      else {
	b=coeff_crs[0][ib] + 
	  coeff_crs[1][ib] * (temper[it]-273.15) + 
	  coeff_crs[2][ib]*(temper[it]-273.15)*(temper[it]-273.15);
	gradient=(a-b)/(lambda_raw[ia]-lambda_raw[ib]);
	crs_out->crs[it][O3][iv] = (gradient*(lambda[iv]-lambda_raw[ib])+b) * fact;
	/*printf("%d %d %f %e\n",it,iv,lambda[iv],crs_out->crs[it][O3][iv]);*/
      }
    }
  }

  return 0;
}

int closest_above(float lambda, float* lambda_raw, int n_crs) 
{
  int iv;

  for (iv=n_crs-1;iv>=0;iv--)  {
    if ((lambda_raw[iv]-lambda) == 0.0) 
      return iv;
    else {
      if ((lambda_raw[iv]-lambda) < 0.0) 
	return iv+1;
    }
  }
  return 0;
}


int closest_below(float lambda, float* lambda_raw, int n_crs) 
{
  int iv;
  for (iv=0; iv<n_crs; iv++) {
    if ( (lambda_raw[iv]-lambda) == 0.0) 
      return iv;
    else  {
      if ((lambda_raw[iv]-lambda) > 0.0) 
	return iv-1;
    }
  }
  return 0;
}
