/*************************************************************************/
/* radiative.c                                                           */
/*                                                                       */
/* Various functions helpful for radiative transfer calculations.        */
/*                                                                       */
/* Author:                                                               */
/*   Arve Kylling                                                  @au@  */
/*   NILU,                                                         @ad@  */
/*   Norway.                                                       @ad@  */
/*                                                                       */
/* ----------------------------------------------------------------------*/
/* Copyright (C) 1997 Arve Kylling                                       */
/*                                                                       */
/* This program is free software; you can redistribute it and/or modify  */
/* it under the terms of the GNU General Public License as published by  */
/* the Free Software Foundation; either version 1, or (at your option)   */
/* any later version.                                                    */
/*                                                                       */
/* This program is distributed in the hope that it will be useful,       */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the          */
/* GNU General Public License for more details.                          */
/*                                                                       */
/* To obtain a copy of the GNU General Public License write to the       */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,    */
/* USA.                                                                  */
/*-----------------------------------------------------------------------*/
/*************************************************************************/

#include <stdio.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <malloc.h>
#include "ascii.h"
#include "radiative.h"
#include "numeric.h"
#include "../tools/uvspec.h"

#define EPSILON 1E-6

/***********************************************************************************/
/* Function: crs_rayleigh                                                 @61_30i@ */
/* Description:                                                                    */
/*    Calculates the Rayleigh scattering cross section according to the formula    */
/*    by Nicolet, M., `On the molecular scattering in the terrestrial atmosphere:  */
/*    an empirical formula for its calculation in the homosphere', Planet. Space   */
/*    Sci., 32, 1467-1468, 1984.                                                   */
/*                                                                                 */
/* Parameters:                                                                     */
/*  float* lambda:         Pointer to ary holding the wavelength in nanometers.    */
/*  int n_lambda:          Number of wavelength points.                            */
/*                                                                                 */
/* Return value:                                                                   */
/*  double pointer to ary holding the Rayleigh scattering cross section as a       */
/*  function of wavelength.                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i61_30@ */
/***********************************************************************************/

int crs_rayleigh(float* lambda, int n_lambda, float ***crs) {
  float ray_const  = 4.02e-28;
  float from_nm_um = 1.e-3;
  float lambda_um=0.0, x;
  int iv;
  
  for (iv=0;iv<n_lambda;iv++) {
    lambda_um = lambda[iv]*from_nm_um;
    if ( lambda_um > 0.55 ) {
      x = 0.04;
    }
    else {
      x = 0.389*lambda_um + 0.09426/lambda_um - 0.3228;
    }
    crs[0][AIR][iv] = ray_const / pow(lambda_um,4+x);
  }
  return 0;
}

/*
*BOD
*DESCRIPTION
* Calculates optical depth, single scattering albedo and
* phase function from absorption and scattering cross sections
* corresponding gas and particulate matter concentrations for 
* one wavelength.
*USAGE
**FORTRAN
*      CALL opt_prop(abs_crs, aer_dtau, aer_gg, aer_ssa,
*     $     dens, dtauc,  ice_dtau, ice_gg, ice_ssa, intgr, nlyr,
*     $     nstr, pmom, ray_crs, ssalb,  wc_dtau, wc_gg, wc_ssa, zd)  
**PERL
*INPUT
**abs_crs
*     real array(temper,trace gas)
*     The absorption cross sections as a function of temper
*     for each absorbing trace gas.
**aer_dtau, aer_gg, aer_ssa
*     real array(layer)
*     The aerosol optical depth, asymmetry factor and single scattering 
*     albedo for each layer.
**dens
*     real array(levels,species)
*     The density of ndens species at nlyr+1 levels. The species index
*     must corresponds to the following trace gases if available. Water
*     vapor, carbon dioxide and nitrogen dioxide are presently not in
*     use and may be set to zero.
***1   Air
***2   ozone (o3)
***3   oxygen (o2)
***4   water vapor (h2o)
***5   carbon dioxide (co2)
***6   nitrogen dioxide (no2)
**ice_dtau, ice_gg, ice_ssa
*     real array(layer)
*     The ice cloud optical depth, asymmetry factor and single scattering 
*     albedo for each layer.
**intgr
*     integer
*     Type of behaviour for integrating density over layer.
***1  Linear behaviour assumed.
***2  Exponential behaviour assumed.
**nlyr
*     integer
*     Number of layers in the model atmosphere.
**nstr
*     integer
*     Number of streams, see disort.
**ray_crs
*     real array
*     The Rayleigh cross section.
**wc_dtau, wc_gg, wc_ssa
*     real array(layer)
*     The water cloud optical depth, asymmetry factor and single scattering 
*     albedo for each layer.
**zd
*     real array(0:*)
*     The altitude in km for levels. 
*OUTPUT
**dtauc
*     real array(*)
*     The optical depth of each layer, see disort.
**pmom
*     real array(streams,layer)
*     The moments of the phase function for each stream and layer.
*     See disort.
**ssalb
*     real array(*)
*     The single scattering albedo of each layer, see disort.
*ERROR_MESSAGES
*ROUTINES_CALLED
*FILES
*     param.inc                Include file
*REFERENCES
*EOD
*/

/***********************************************************************************/
/* Function: optical_properties                                           @61_30i@ */
/* Description:                                                                    */
/*    Calculates the Rayleigh scattering cross section according to the formula    */
/*    by Nicolet, M., `On the molecular scattering in the terrestrial atmosphere:  */
/*    an empirical formula for its calculation in the homosphere', Planet. Space   */
/*    Sci., 32, 1467-1468, 1984.                                                   */
/*                                                                                 */
/* Parameters:                                                                     */
/*  float* lambda:         Pointer to ary holding the wavelength in nanometers.    */
/*  int n_lambda:          Number of wavelength points.                            */
/*                                                                                 */
/* Return value:                                                                   */
/*  double pointer to ary holding the Rayleigh scattering cross section as a       */
/*  function of wavelength.                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i61_30@ */
/***********************************************************************************/

int optical_properties( int iv, atm_inp_struct *atm_inp, atm_out_struct *atm_out,
			crs_out_struct *crs_out, output_struct *output, rte_struct *rte) {
  static int first=1;
  int k, lu;
  int status=0;
  float bscar=0, babs=0, babsa=0, babsd=0, babsi=0, babso=0,
    bext=0, bsca=0, bscaa=0, bscad=0, bscai=0, deltaz=0, ga=0, gd=0, gi=0;
  float sumsca=0;

  if ( first ) {
    first = 0;
    
    /* Calculate column density of air (total density of the atmosphere). */
    /* Used to calculate Rayleigh optical depth. */
    
    atm_out->cdento  = (float *) calloc (atm_out->nlyr, sizeof(float));
    output->dtauc    = (float *) calloc (atm_out->nlyr, sizeof(float));
    output->ssalb    = (float *) calloc (atm_out->nlyr, sizeof(float));
    output->dtair    = (float *) calloc (atm_out->nlev, sizeof(float));
    output->dto3     = (float *) calloc (atm_out->nlev, sizeof(float));
    output->dtaer    = (float *) calloc (atm_out->nlev, sizeof(float));
    output->dtcld    = (float *) calloc (atm_out->nlev, sizeof(float));
    output->dtice    = (float *) calloc (atm_out->nlev, sizeof(float));
    output->omaer    = (float *) calloc (atm_out->nlev, sizeof(float));
    output->omcld    = (float *) calloc (atm_out->nlev, sizeof(float));
    output->omice    = (float *) calloc (atm_out->nlev, sizeof(float));
    output->fgaer    = (float *) calloc (atm_out->nlev, sizeof(float));
    output->fgcld    = (float *) calloc (atm_out->nlev, sizeof(float));
    output->fgice    = (float *) calloc (atm_out->nlev, sizeof(float));
    atm_out->zmc     = (float *) calloc (atm_out->nlev, sizeof(float));
    status           = ASCII_calloc_float(&output->pmom, atm_out->nlyr, rte->nstr+1);
    for (lu=0;lu<atm_out->nlyr;lu++) {
      deltaz = ( atm_out->zd[lu]- atm_out->zd[lu+1] )*1.e+5;
	/* The factor 1.e+5 converts z from km to cm.*/
      if (atm_out->dens[AIR][lu] > 0 && atm_out->dens[AIR][lu+1] > 0 ) {
	if (fabs(1.-atm_out->dens[AIR][lu+1]/atm_out->dens[AIR][lu]) > EPSILON )  {
	  atm_out->cdento[lu] = 1. /(log(atm_out->dens[AIR][lu+1]/atm_out->dens[AIR][lu])) * 
	    (atm_out->dens[AIR][lu+1] - atm_out->dens[AIR][lu]) * deltaz ;
	}
	else {
	  atm_out->cdento[lu] = 0.5*fabs((atm_out->dens[AIR][lu]+
					  atm_out->dens[AIR][lu+1])*deltaz);
	}
      }
      else {
	atm_out->cdento[lu] = 0.5*fabs((atm_out->dens[AIR][lu]+
					atm_out->dens[AIR][lu+1])*deltaz);
      }      
    }
  }
  
  for (lu=0;lu<atm_out->nlyr;lu++) {
    bscar = atm_out->cdento[lu]*crs_out->crs[0][AIR][iv];
    deltaz= ( atm_out->zd[lu]- atm_out->zd[lu+1] )*1.e+5;
    babso = 0.5*deltaz*( atm_out->dens[O3][lu]*crs_out->crs[lu][O3][iv]
			 +  atm_out->dens[O3][lu+1]*crs_out->crs[lu+1][O3][iv]);
	/* The factor 1.e+5 converts z from km to cm.*/

    /*    printf("opt_prop %d %f %e %f %f %e %e\n", lu, deltaz, 
	   atm_out->cdento[lu], bscar, babso, 
	   atm_out->dens[O3][lu+1],crs_out->crs[lu+1][O3][iv]);*/

    bscaa  = output->aer.ssa[iv][lu]*output->aer.dtau[iv][lu];
    bscad  = output->wc.ssa[iv][lu]*output->wc.dtau[iv][lu];
    bscai  = output->ice.ssa[iv][lu]*output->ice.dtau[iv][lu];

    ga     = output->aer.gg[iv][lu];
    gd     = output->wc.gg[iv][lu];
    gi     = output->ice.gg[iv][lu];

    /*  Phase function */
    sumsca =  bscar + bscad +  bscaa + bscai;
    if ( sumsca < FLT_MIN ) { sumsca = FLT_MIN; }

    output->pmom[lu][0] = 1.0;
    for (k=1;k<=rte->nstr;k++) {
      output->pmom[lu][k] = (bscad*pow(gd,k) + bscaa*pow(ga,k) + bscai*pow(gi,k)) / sumsca;
      if ( k == 2 ) {
	output->pmom[lu][k] = output->pmom[lu][k]+bscar/(sumsca*10.);
      }
      /*      printf("%d %d %f %f %f %f %f %f %f %f\n", k, lu, 
	     output->pmom[lu][k], 
	     output->aer.gg[iv][lu],sumsca,
	     pow(bscaa*ga,k), bscaa, ga,
	     output->aer.ssa[iv][lu],output->aer.dtau[iv][lu]);*/
    }

    /*      Absorption coefficient        */
    babsa  = (1.-output->aer.ssa[iv][lu])*output->aer.dtau[iv][lu];
    babsd  = (1.-output->wc.ssa[iv][lu])*output->wc.dtau[iv][lu];
    babsi  = (1.-output->ice.ssa[iv][lu])*output->ice.dtau[iv][lu];
    babs = babsa + babso + babsd + babsi;
    if (babs > FLT_MAX) {babs = FLT_MAX;}
    
    /*     Scattering coefficient   */
    bsca = bscar + bscad + bscaa + bscai;
    if (bsca > FLT_MAX) {bsca = FLT_MAX;}
    
    /*  Extinction coefficient    */
    bext = babs + bsca;
    if (bext > FLT_MAX) {bext = FLT_MAX;}

    /* Optical properties for the mc model */

    output->dtair[atm_out->nlyr-1-lu]   = bscar;
    output->dto3 [atm_out->nlyr-1-lu]   = babso;
    output->dtaer[atm_out->nlyr-1-lu]   = babsa + bscaa;
    output->dtcld[atm_out->nlyr-1-lu]   = babsd + bscad;
    output->dtice[atm_out->nlyr-1-lu]   = babsi + bscai;

    output->omaer[atm_out->nlyr-1-lu]   = output->aer.ssa[iv][lu];
    output->omcld[atm_out->nlyr-1-lu]   = output->wc.ssa[iv][lu];
    output->omice[atm_out->nlyr-1-lu]   = output->ice.ssa[iv][lu];
    
    /* output->omaer[atm_out->nlyr-1-lu]   = output->aer.gg[iv][lu]; */
    /* output->omcld[atm_out->nlyr-1-lu]   = output->wc.gg[iv][lu];  */
    /* output->omice[atm_out->nlyr-1-lu]   = output->ice.gg[iv][lu]; */
    
    output->fgaer[atm_out->nlyr-1-lu]   = output->aer.gg[iv][lu];
    output->fgcld[atm_out->nlyr-1-lu]   = output->wc.gg[iv][lu];
    output->fgice[atm_out->nlyr-1-lu]   = output->ice.gg[iv][lu];
    
    /*      Set optical depth and single scattering albedo */
    if (bext < FLT_MIN) {bext = FLT_MIN;}
    output->dtauc[lu] = bext;
    output->ssalb[lu] = bsca/bext;
    if ( 1.0-output->ssalb[lu] < FLT_MIN ) {output->ssalb[lu] = 1.0 - FLT_MIN; }
  }
  output->dtair[atm_out->nlyr]   = 0.0;
  output->dto3 [atm_out->nlyr]   = 0.0;
  output->dtaer[atm_out->nlyr]   = 0.0; 
  output->dtcld[atm_out->nlyr]   = 0.0;
  output->dtice[atm_out->nlyr]   = 0.0;
                                   
  output->omaer[atm_out->nlyr]   = 0.0;
  output->omcld[atm_out->nlyr]   = 0.0;
  output->omice[atm_out->nlyr]   = 0.0;
                                   
  output->omaer[atm_out->nlyr]   = 0.0;
  output->omcld[atm_out->nlyr]   = 0.0;
  output->omice[atm_out->nlyr]   = 0.0;
  for (lu=0;lu<=atm_out->nlyr;lu++) {
    atm_out->zmc[atm_out->nlyr-lu]     = atm_out->zd[lu];
  }
  return 0;
}

int set_out (float* dtauc, int nlyr, int nzout, 
	     float* utau, float* zd, float* zout ) {
/* * @20i@ */
/* * Function: setout */
/* * Description: */
/* * Linearly interpolate the cumulative optical depth (calculated from  */
/* * @code{dtauc} to get approximate @code{utau} corresponding to altitudes  */
/* * @code{zout}. */
/* * */
/* * Parameters: */
/* *    REAL(*) dtauc:  Optical depth of each layer. (input) */
/* *    INTEGER nlyr:   Number of atmospheric layers. (input). */
/* *    INTEGER ntau:   Number of output altitudes. (input). */
/* *    REAL(*) zd:     The altitude of each level in km. @code{zd(nlyr)} is the  */
/* *                    bottom of the atmosphere. nlyr+1 elements. (input). */
/* *    REAL(*) zout:   The altitude of @code{ntau} output altitudes. (input). */
/* *    REAL(*) utau:   The optical depth at @code{ntau} altitudes. (output). */
/* * */
/* * Return value: */
/* * Example: */
/* * See @file{uvspec.c}. */
/* * Files:  */
/* * @file{param.inc} with parameter @code{maxcly} must be present in the  */
/* * same directory as @file{wcloud.f}. */
/* * Known bugs: */
/* * Author:  Arve Kylling     */
/* *      */
/* * @i20@ */
/* * */
/*       SUBROUTINE setout( dtauc, nlyr, ntau, utau, zd, zout) */
/* * */
/*       IMPLICIT LOGICAL (A-Z)    ! To nearly get strong typing */
/*       INCLUDE 'param.inc' */
/* * */
/*       INTEGER itype, lc, lu, nlyr, ntau */
/*       LOGICAL usrtau */
/*       REAL dtauc(*), hh, utau(*), zd(0:*), zout(*), utauout */
/*       REAL dummy(maxcly+1) */
/* * */
  double* cumtau;
  int i, lu, status;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *x=NULL, *y=NULL;
  double ynew;

  cumtau = (double *) calloc (nlyr+1, sizeof(double));
  cumtau[0] = 0;
  for (lu=1;lu<=nlyr;lu++) {
    cumtau[lu] = cumtau[lu-1] + (double)dtauc[lu-1]; 
  }

  x = (double *) calloc (nlyr+1, sizeof(double));
  y = (double *) calloc (nlyr+1, sizeof(double));
  for (i=0;i<=nlyr;i++) {
    x[nlyr-i] = (double) zd[i];
    y[nlyr-i] = (double) cumtau[i];
  }
  status = linear_coeffc (x, y, nlyr+1, &a0, &a1, &a2, &a3);
  if (status!=0)  {
    fprintf (stderr, "set_out: sorry cannot do linear interpolation\n");
    fprintf (stderr, "linear_coeffc() returned status %d\n", status);
    return status;
  }
  for (i=0; i<nzout; i++)  {
    status = calc_splined_value ((double) zout[i], &ynew, x, nlyr+1, a0, a1, a2, a3);
    if (status!=0)  {
      fprintf (stderr, "set_out: sorry cannot do linear interpolation\n");
      fprintf (stderr, "linear_coeffc() returned status %d\n", status);
      return status;
    }
    utau[i] = (float) ynew;
    /*    printf("ynew %d %f\n",i,ynew);*/
  }
  free(a0);
  free(a1);
  free(a2);
  free(a3);
  free(x);
  free(y);
  free(cumtau);
  return 0;
}

