#
# Copyright (C) 1997 Arve Kylling
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# To obtain a copy of the GNU General Public License write to the
# Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
# USA.

#-------------------------------------------------------------------------

use lib "../tools";
use UVTools;
use Getopt::Long;
use Carp;
use strict;

my $answers      = 0;

my $ret = GetOptions(
		  "--answers",\$answers);
if ($ret==0) { exit; }

my $failure=0;
my $maxdiff  = 0.005; # Accepted absolute error between test cases and
                      # answers

printf stderr "Running various libRadtran tests. This may take some time....\n";

printf stderr "make_slitfunction test\n";
$failure=0;
$failure+=make_slitfunction_test($answers,$maxdiff);
report_results($answers,$failure,"make_slitfunction");

printf stderr "Some uvspec tests\n";
$failure=0;
$failure+=uvspec_test_clear($answers,$maxdiff);
$failure+=uvspec_test_aerosol($answers,$maxdiff);
$failure+=uvspec_test_wc($answers,$maxdiff);
$failure+=uvspec_test_radiances($answers,$maxdiff);
$failure+=uvspec_test_twostr($answers,$maxdiff);
$failure+=uvspec_test_sdisort($answers,$maxdiff);
$maxdiff = 0.02;  # Increase accepted error for mc model
$failure+=uvspec_test_mc($answers,$maxdiff);
$maxdiff = 0.005;
$failure+=atm_reflectivity_test($answers,$maxdiff); 
report_results($answers,$failure,"uvspec");

printf stderr "Calc_J test\n";
$failure=0;
$failure+=calc_j_test($answers,$maxdiff);
report_results($answers,$failure,"Calc_J");

exit;

sub atm_reflectivity_test {
    my $answers = shift @_;
    my $failure=0;
    my $pi = 4*atan2(1,1);

    # Create solar flux file with unit flux
    open(FP,">TMP.SUN");
    for(my $i=300;$i<=340;$i++) {printf FP "%f 1.0\n", $i;}
    close(FP);

    my $test = "atmospheric reflectivity"; 
    my $inp = "TMP.INP";
    my $out = "TMP.OUT1";
    my $exinp  = "../examples/UVSPEC_REF1.INP";
    printf stderr "$test ...";
    system("cat  $exinp > $inp");
    my $ret = system("$UVTools::uvspec < $inp > $out");
    if ($ret != 0) {carp "uvspec failed"; return 1;} 

    $exinp  = "../examples/UVSPEC_REF2.INP";
    system("cat  $exinp > $inp");
    $out = "TMP.OUT2";
    my $ret = system("$UVTools::uvspec < $inp > $out");
    if ($ret != 0) {carp "uvspec failed"; return 1;} 

    $exinp  = "../examples/UVSPEC_REF3.INP";
    system("cat  $exinp > $inp");
    $out = "TMP.OUT3";
    my $ret = system("$UVTools::uvspec < $inp > $out");
    if ($ret != 0) {carp "uvspec failed"; return 1;} 

    open(FP1,"TMP.OUT1");
    open(FP2,"TMP.OUT2");
    open(FP3,"TMP.OUT3");
    while (my $line1 = <FP1>) {
	my $line2 = <FP2>;
	my $line3 = <FP3>;
	$line1 =~ s/^[ ]+//; $line2 =~ s/^[ ]+//; $line3 =~ s/^[ ]+//;
	my ($wvn1,$dir1,$dn1,$up1) = split(/ +/,$line1);
	my ($wvn2,$dir2,$dn2,$up2) = split(/ +/,$line2);
	my ($wvn3,$dir3,$dn3,$up3) = split(/ +/,$line3);
	$up1 = $up1/$pi; # Divide by pi to get reflectivity
	if (abs(1-$up1/(1-($dir2+$dn2)/($dir3+$dn3)))>0.0001) {
	    $failure++
	}
#	printf "%f %f %f %f %d\n", $wvn1, $up1, 1-($dir2+$dn2)/($dir3+$dn3), 
#	$up1/(1-($dir2+$dn2)/($dir3+$dn3)), $failure;
    }
    close(FP1);
    close(FP2);
    close(FP3);
    if ($failure==0) {  printf stderr " ok\n"; }
    else { printf stderr "Found %d serious differences for $test test.\n", $failure;}

    `rm -f TMP.OUT1 TMP.OUT2 TMP.OUT3 TMP.INP TMP.SUN`;
    $failure;
}
   
sub make_slitfunction_test {
    my $answers = shift @_;
    my $maxdiff = shift @_;
    my $failure = 0;
    my $diffout= "TMP_DIFF.OUT";
    my $test = "make_slitfunction";
    my $nox = "";
    printf stderr "$test ...";
    my $inp = "";
    my $out = "TMP.OUT";
    my $exout  = "../examples/TRI_SLIT.DAT";
    my $cmd = "../tools/make_slitfunction -f 0.75 -r 0.01 -t 1";
    open(FP,">".$out) || carp "Could not open file $out";
    printf FP "# Slit function made by executing:\n";
    printf FP "#    $cmd\n";
    close(FP);
    my $ret = system("$cmd >> $out");
    if ($ret != 0) {carp "make_slitfunction failed"; return 1;} 
    return report_differences($answers,$out,$exout,$inp,$test,$nox,$maxdiff);
}

sub uvspec_test_clear {
    my $answers = shift @_;
    my $maxdiff = shift @_;
    my $uvspec_failure=0;
    my $test = "disort clear sky"; 
    my $inp = "TMP.INP";
    my $out = "TMP.OUT";
    my $exinp  = "../examples/UVSPEC_CLEAR.INP";
    my $exout  = "../examples/UVSPEC_CLEAR.OUT";
    $uvspec_failure+=uvspec_run_test($inp, $out, $exinp, $exout, $test, $answers,$maxdiff);
    $uvspec_failure;
}
   
sub uvspec_test_aerosol {
    my $answers = shift @_;
    my $maxdiff = shift @_;
    my $uvspec_failure=0;
    my $test = "disort aerosol"; 
    my $inp = "TMP.INP";
    my $out = "TMP.OUT";
    my $exinp  = "../examples/UVSPEC_CLEAR.INP ../examples/UVSPEC_AEROSOL.INP";
    my $exout  = "../examples/UVSPEC_AEROSOL.OUT";
    $uvspec_failure+=uvspec_run_test($inp, $out, $exinp, $exout, $test, $answers,$maxdiff);
    $uvspec_failure;
}
   
sub uvspec_test_wc {
    my $answers = shift @_;
    my $maxdiff = shift @_;
    my $uvspec_failure=0;
    my $test = "disort water cloud"; 
    my $inp = "TMP.INP";
    my $out = "TMP.OUT";
    my $exinp  = "../examples/UVSPEC_CLEAR.INP ../examples/UVSPEC_WC.INP";
    my $exout  = "../examples/UVSPEC_WC.OUT";
    $uvspec_failure+=uvspec_run_test($inp, $out, $exinp, $exout, $test, $answers,$maxdiff);
    $uvspec_failure;
}
   
sub uvspec_test_radiances {
    my $answers = shift @_;
    my $maxdiff = shift @_;
    my $uvspec_failure=0;
    my $test = "disort radiances"; 
    my $inp = "TMP.INP";
    my $out = "TMP.OUT";
    my $exinp  = "../examples/UVSPEC_CLEAR.INP ../examples/UVSPEC_AEROSOL.INP ../examples/UVSPEC_RADIANCES.INP";
    my $exout  = "../examples/UVSPEC_RADIANCES.OUT";
    $uvspec_failure+=uvspec_run_test($inp, $out, $exinp, $exout, $test, $answers,$maxdiff);
    $uvspec_failure;
}
   
sub uvspec_test_twostr {
    my $answers = shift @_;
    my $maxdiff = shift @_;
    my $uvspec_failure=0;
    my $test = "twostr aerosol and water cloud"; 
    my $inp = "TMP.INP";
    my $out = "TMP.OUT";
    `cp ../examples/UVSPEC_CLEAR.INP TMPA.INP`;
    `sed s/disort/twostr/ TMPA.INP > TMPB.INP; rm -f TMPA.INP`;
    my $exinp  = "TMPB.INP ../examples/UVSPEC_AEROSOL.INP ../examples/UVSPEC_WC.INP";
    my $exout  = "../examples/UVSPEC_TWOSTR.OUT";
    $uvspec_failure+=uvspec_run_test($inp, $out, $exinp, $exout, $test, $answers,$maxdiff);
    `rm -f TMPB.INP`;
    $uvspec_failure;
}

sub uvspec_test_sdisort {
    my $answers = shift @_;
    my $maxdiff = shift @_;
    my $uvspec_failure=0;
    my $test = "sdisort aerosol"; 
    my $inp = "TMP.INP";
    my $out = "TMP.OUT";
    `cp ../examples/UVSPEC_CLEAR.INP TMPA.INP`;
    `sed s/disort/sdisort/ TMPA.INP > TMPB.INP; rm -f TMPA.INP`;
    my $exinp  = "TMPB.INP ../examples/UVSPEC_AEROSOL.INP";
    my $exout  = "../examples/UVSPEC_SDISORT.OUT";
    $uvspec_failure+=uvspec_run_test($inp, $out, $exinp, $exout, $test, $answers,$maxdiff);
    `rm -f TMPB.INP`;
    $uvspec_failure;
}
   
sub uvspec_test_mc {
    my $answers = shift @_;
    my $maxdiff = shift @_;
    my $uvspec_failure=0;
    my $test = "mc water cloud"; 
    my $inp = "TMP.INP";
    my $out = "TMP.OUT";
    if (-e "../src_c/mc.c")  {
	`cp ../examples/UVSPEC_MC_ALB.INP ./albedo2D.dat`;
	`cp ../examples/UVSPEC_MC_WC.INP ./wcloud3D.dat`;
	`cp ../examples/UVSPEC_MC_SAMPLE.INP ./sample2D.dat`;
	`cp ../examples/UVSPEC_MC_ELEV.INP ./elevation2D.dat`;
	my $exinp  = "../examples/UVSPEC_MC.INP";
	my $exout  = "../examples/UVSPEC_MC.OUT";
	$uvspec_failure+=uvspec_run_test($inp, $out, $exinp, $exout, $test, $answers,$maxdiff);
    }
    `rm -f albedo2D.dat elevation2D.dat mc.avg. mc.bin mc.rad sample2D.dat wcloud3D.dat`;
    $uvspec_failure;
}

sub uvspec_run_test {
    my ($inp, $out, $exinp, $exout, $test, $answers, $maxdiff) = @_;
    printf stderr "$test ...";
    system("cat  $exinp > $inp");
    my $ret = system("$UVTools::uvspec < $inp > $out");
    if ($ret != 0) {carp "uvspec failed"; return 1;} 
    my $nox = "";
    return report_differences($answers,$out,$exout,$inp,$test,$nox,$maxdiff);
}

sub calc_j_test {
    my $aswers=shift @_;
    my $maxdiff = shift @_;
    my $failure=0;
    my $test = "j values"; 
    printf stderr "$test ...";
    my $inp = "TMP.INP";
    my $out = "TMP.OUT";
    my $exinp  = "../examples/J_UVSPEC.INP";
    my $exout  = "../examples/J.OUT";
    $failure+=calc_j_run_test($inp, $out, $exinp, $exout, $test, $answers, $maxdiff);
    $failure;
}
   
sub calc_j_run_test {
    my ($inp, $out, $exinp, $exout, $test, $answers, $maxdiff) = @_;
    my $diffout= "TMP_DIFF.OUT";
    system("cat  $exinp > $inp");
    my $uvspec_out = "../examples/J_UVSPEC.OUT";
    my $ret =  system("$UVTools::uvspec < $inp > $uvspec_out");
    if ($ret != 0) {carp "uvspec failed";return 1;} 
    my $temp = 300.0;
    my $string = "'self_surface_temperature(K) $temp'";
    system("echo $string >> $uvspec_out");
    $ret = system("perl ../tools/Calc_J.pl $uvspec_out > $out");
    if ($ret != 0) {carp "Calc_J.pl failed"; return  1;} 
    my $nox = "--nox";
    return report_differences($answers,$out,$exout,$inp,$test,$nox,$maxdiff);
}

sub report_differences {
    my $answers = shift @_;
    my $out     = shift @_;
    my $exout   = shift @_;
    my $inp     = shift @_;
    my $test    = shift @_;
    my $nox     = shift @_;
    my $maxdiff   = shift @_;
    my $diffout = "TMP_DIFF.OUT";
    
    if ( $answers == 1 ) {
	system("mv $out $exout");
	if ( -e $inp ) {system("rm -f $inp");}
	printf stderr " $exout replaced\n";
    }
    else {
	if (!-e $exout ) {croak "Example output file $exout not found";}
	if (!-e $out ) {croak "Test  output file $out not found";}
	my $ret = system("perl ../tools/ndiff $nox --quiet --limit 0.01 --maxdiff $maxdiff $exout $out > $diffout");
	if ($ret != 0) { carp "ndiff failed"; return 1; }
	open(FILE,$diffout);
	my $ndiffs = <FILE>;
	close(FILE);
	chop $ndiffs;
	if ($ndiffs==0) {
	    printf stderr " ok\n";
	}
	else
	{
	    printf stderr "Found %d serious differences for $test test.\n", $ndiffs;
	}
	system("rm -f $diffout $inp $out");
	return $ndiffs;
    }
}

sub report_results{
    
    my $answers = shift @_;
    my $failure = shift @_;;
    my $test    = shift @_;
    if ( $answers != 1 ) {
    if ($failure) {
	printf stderr "$failure of the $test tests failed.\n";
    }
    else {
	printf stderr "All $test tests succeeded.\n";
    }
}
    else {
	printf stderr "All *.OUT files replaced.\n";
    }
}


# Local Variables:
# mode: Perl
# End:
