/************************************************************************/
/*                                                                      */
/* Copyright (C) 1997 Arve Kylling                                      */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*                                                                      */
/************************************************************************/

#include <stdio.h>							    
#include <stdlib.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/numeric.h"


/* prototypes of internal functions */
static int cnvlv (double *x_spec, float  *y_spec, int n_spec,
		  double *x_slit, double *y_slit, int n_slit);


/**************************************************************/
/* Setup the user-defined wavelength grid.                    */
/**************************************************************/

int setup_wlgrid (input_struct input, output_struct *output)
{     
  int status=0, rows=0, min_columns=0, max_columns=0;
  float **wl_data=NULL;


  if ( strlen(input.transmittance_wl_file) > 0 ) {
    
    status = ASCII_file2float (input.transmittance_wl_file, 
			       &rows, &max_columns, &min_columns, 
			       &wl_data);
    if (status!=0) {
      fprintf (stderr, " ... error %d opening file %s\n", status, input.transmittance_wl_file);
      return status;
    }

    if (max_columns!=min_columns) {
      fprintf (stderr, " !! ATTENTION !! Inconsistent number of columns\n");
      fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
    }  

    if (min_columns!=1) {
      fprintf (stderr, " ... ending, too few columns\n");
      return -1;
    }

    output->wl.nlambda_t = rows;
    output->wl.lambda_t  = ASCII_column_float(wl_data, output->wl.nlambda_t, 0 );
    
    status = ASCII_free_float(wl_data, rows);
  }
  else 
    status = set_transmittance_wl_grid(&input.wl, &output->wl);
  
  return 0;
}  






/**************************************************************/
/* Interpolate from one wavelength grid to another, either    */
/* with natural cubic splines (linear=0) or linear (linear=1) */ 
/**************************************************************/

int arb_wvn( int n_old_x, float *old_x, float *old_y,  
	     int n_new_x, float *new_x, float *new_y, int linear)
{
  int i=0, status=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *x=NULL, *y=NULL;
  double ynew=0.0;

  x = (double *) calloc (n_old_x, sizeof(double));
  y = (double *) calloc (n_old_x, sizeof(double));

  for (i=0;i<n_old_x;i++) {
    x[i] = (double) old_x[i];
    y[i] = (double) old_y[i];
  }

  if (linear) {
    status = linear_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do linear interpolation\n");
      fprintf (stderr, "linear_coeffc() returned status %d\n", status);
      return status;
    }
  }
  else {
    status = spline_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do spline interpolation\n");
      fprintf (stderr, "spline_coeffc() returned status %d\n", status);
      return status;
    }
  }

  for (i=0; i<n_new_x; i++)  {
    status = calc_splined_value ((double) new_x[i], &ynew, x, n_old_x, a0, a1, a2, a3);
    new_y[i] = (float) ynew;
  }

  free(a0); free(a1); free(a2); free(a3);
  free(x);
  free(y);

  return 0;
}




/**************************************************************/
/* Define the internal wavelength grid for the transmittance  */
/* calculation.                                               */
/**************************************************************/

int set_transmittance_wl_grid(wl_inp_struct *wl_inp, wl_out_struct *wl_out)
{
  int iv=0;
  float lambda=0.0, wvn_step_t=0.0;
  
  /* Calculate number of wavelengths needed for transmittance calculation */

  lambda = wl_inp->wl_start;
  iv = 1;
  while (lambda < wl_inp->wl_end) {
    if ( lambda < 350.0 )
      wvn_step_t  = 0.5;
    else
      wvn_step_t  = 1.0;

    lambda += wvn_step_t;
    iv++;
  }

  wl_out->nlambda_t = iv;
  wl_out->lambda_t = (float *) calloc (wl_out->nlambda_t, sizeof(float));


  /* Set wavelengths for transmittance calculation */
  lambda = wl_inp->wl_start;
  for (iv=0; iv<wl_out->nlambda_t; iv++) {
    wl_out->lambda_t[iv] = lambda;
    if ( lambda < 350.0 )
      wvn_step_t  = 0.5;
    else
      wvn_step_t  = 1.0;

    lambda += wvn_step_t;
  }

  return 0;
}



/**************************************************************/
/* Convolve the spectrum with a user-defined slit function    */
/**************************************************************/

int convolve (input_struct input, output_struct *output)
{
  int iv=0, iu=0, j=0, lev=0;
  int status=0;
  double *x_data=NULL;

  int n_slit=0;
  double *x_slit=NULL, *y_slit=NULL;


  /* read slit function from file */
  status = read_2c_file (input.slit_function_filename, &x_slit, &y_slit, &n_slit); 
  if (status!=0)  { 
    fprintf (stderr, " ... error %d reading file %s\n", status, input.slit_function_filename); 
    return status; 
  } 

  x_data = calloc (output->wl.nlambda_h, sizeof(double)); 

  for (iv=0;iv<output->wl.nlambda_h;iv++)
    x_data[iv] = (double) output->wl.lambda_h[iv];
  
  for (lev=0;lev<input.atm.nzout;lev++) {

    /* Convolve rfldir, rfldn, flup and uavg for all cases */

    /* rfldir */
    status = cnvlv (x_data, output->rfldir[lev], output->wl.nlambda_h,
		    x_slit, y_slit, n_slit);
    if (status!=0)  { 
      fprintf (stderr, " ... error %d convoluting rfldir\n", status); 
      return status; 
    } 

    /* rfldn */
    status = cnvlv (x_data, output->rfldn[lev], output->wl.nlambda_h,
		    x_slit, y_slit, n_slit);
    if (status!=0)  { 
      fprintf (stderr, " ... error %d convoluting rfldn\n", status); 
      return status; 
    } 

    /* flup */
    status = cnvlv (x_data, output->flup[lev], output->wl.nlambda_h,
		    x_slit, y_slit, n_slit);
    if (status!=0)  { 
      fprintf (stderr, " ... error %d convoluting flup\n", status); 
      return status; 
    } 

    /* uavg */
    status = cnvlv (x_data, output->uavg[lev], output->wl.nlambda_h,
		    x_slit, y_slit, n_slit);
    if (status!=0)  { 
      fprintf (stderr, " ... error %d convoluting uavg\n", status); 
      return status; 
    } 


    if (input.rte.solver != TWOSTR) {
      
      /* Convolve uavgso, uavgdn, uavgup */
      
      /* uavgso */
      status = cnvlv (x_data, output->uavgso[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, " ... error %d convoluting uavgso\n", status); 
	return status; 
      } 

      /* uavgdn */
      status = cnvlv (x_data, output->uavgdn[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, " ... error %d convoluting uavgdn\n", status); 
	return status; 
      } 

      /* uavgup */
      status = cnvlv (x_data, output->uavgup[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, " ... error %d convoluting uavgup\n", status); 
	return status; 
      } 


      if ( input.rte.numu > 0 ) {
      
	/* Convolve u0u */
	for (iu=0; iu<input.rte.maxumu; iu++) {
	  status = cnvlv (x_data, output->u0u[lev][iu], output->wl.nlambda_h,
			  x_slit, y_slit, n_slit);
	  if (status!=0)  { 
	    fprintf (stderr, " ... error %d convoluting u0u\n", status); 
	    return status; 
	  } 
	}
	
	if ( output->print.phi > 0 ) {
	  
	  /* Convolve uu */
	  for (j=0; j<input.rte.nphi; j++) {
	    for (iu=0; iu<input.rte.maxumu; iu++) {
	      status = cnvlv (x_data, output->uu[lev][j][iu], output->wl.nlambda_h,
			      x_slit, y_slit, n_slit);
	      if (status!=0)  { 
		fprintf (stderr, " ... error %d convoluting u0u\n", status); 
		return status; 
	      } 
	    }
	  }
	}
      }
    }
  }

  free(x_data);  

  return 0;
}



/**************************************************************/
/* Convolve a spectrum with a slit function.                  */
/* Internal function, used by convolve().                     */
/**************************************************************/

static int cnvlv (double *x_spec, float  *y_spec, int n_spec,
		  double *x_slit, double *y_slit, int n_slit) 
{
  int iv=0, status=0;
  double *y_data = calloc (n_spec, sizeof(double));
  double *tmp=0;

  for (iv=0; iv<n_spec; iv++) 
    y_data[iv] = (double) y_spec[iv];
  
  status = int_convolute (x_spec, y_data, n_spec, x_slit, y_slit, n_slit, &tmp);
  if (status!=0) 
    return status; 
  
  for (iv=0; iv<n_spec; iv++) 
    y_spec[iv] = (float) tmp[iv];
  
  free(y_data);
  free(tmp);

  return 0;
}


/**************************************************************/
/* Interpolate the hires spectrum to a user-defined           */
/* wavelength grid.                                           */
/**************************************************************/

int spline_interpolate(input_struct input, output_struct *output)
{
  int iu=0, iv=0, j=0, lev=0;
  int status=0;
  double *x_user=NULL;
  int linear=0;


  if (strlen(input.spline_filename) > 0) {

    /* read file with user x values */
    status = read_1c_file (input.spline_filename, &x_user, &output->wl.nlambda_s);
    if (status!=0)  {
      fprintf (stderr, "spline_interpolate: error %d reading file %s\n", 
	       status, input.spline_filename);
      return (status);
    }

    output->wl.lambda_s = (float *) calloc (output->wl.nlambda_s, sizeof(float));

    for (iv=0; iv<output->wl.nlambda_s; iv++)
      output->wl.lambda_s[iv] = (float) x_user[iv];

    /* Check wavelength ranges */
    status=0;
    if (input.wl.wl_start > output->wl.lambda_s[0])  {
      fprintf (stderr, "Spline lambda %f is smaller than wvn lambda %f\n",
	       output->wl.lambda_s[0], input.wl.wl_start);
      status--;
    }

    if (input.wl.wl_end < output->wl.lambda_s[output->wl.nlambda_s-1])  {
      fprintf (stderr, "Spline lambda %f is greater than wvn lambda %f\n",
	       output->wl.lambda_s[output->wl.nlambda_s-1], input.wl.wl_end);
      status--;
    }

    if (status!=0) 
      return status;
  }
  else {
    output->wl.nlambda_s = 
      (int) (input.spline_lambda_1-input.spline_lambda_0)/input.spline_lambda_step+1;

    output->wl.lambda_s = (float *) calloc (output->wl.nlambda_s, sizeof(float));

    for (iv=0;iv<output->wl.nlambda_s;iv++)
      output->wl.lambda_s[iv] = input.spline_lambda_0 + (float) iv * input.spline_lambda_step;
  }
  

  /* Spline interpolate to given wavelengths */
  for (lev=0; lev<input.atm.nzout; lev++) {

    /* Spline interpolate rfldir, rfldn, flup and uavg for all cases */
	  
    /* rfldir */
    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldir[lev],  
		     output->wl.nlambda_s, output->wl.lambda_s, output->rfldir[lev], linear);
    if (status!=0) {
      fprintf (stderr, " ... error %d interpolating rfldir\n", status);
      return status;
    }

    /* rfldn */
    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldn[lev],  
		     output->wl.nlambda_s, output->wl.lambda_s, output->rfldn[lev], linear);
    if (status!=0) {
      fprintf (stderr, " ... error %d interpolating rfldn\n", status);
      return status;
    }

    /* flup */
    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->flup[lev],  
		     output->wl.nlambda_s, output->wl.lambda_s, output->flup[lev], linear);
    if (status!=0) {
      fprintf (stderr, " ... error %d interpolating flup\n", status);
      return status;
    }

    /* uavg */
    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavg[lev],  
		     output->wl.nlambda_s, output->wl.lambda_s, output->uavg[lev], linear);
    if (status!=0) {
      fprintf (stderr, " ... error %d interpolating uavg\n", status);
      return status;
    }

    if (input.rte.solver != TWOSTR) {

      /* Spline interpolate uavgso, uavgdn, uavgup */
      
      /* uavgso */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgso[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->uavgso[lev], linear);
      if (status!=0) {
	fprintf (stderr, " ... error %d interpolating uavgso\n", status);
	return status;
      }

      /* uavgdn */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgdn[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->uavgdn[lev], linear);
      if (status!=0) {
	fprintf (stderr, " ... error %d interpolating uavgdn\n", status);
	return status;
      }

      /* uavgup */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgup[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->uavgup[lev], linear);
      if (status!=0) {
	fprintf (stderr, " ... error %d interpolating uavgup\n", status);
	return status;
      }

      if (input.rte.numu > 0) {

	/* Spline interpolate u0u */
	for (iu=0;iu<input.rte.maxumu;iu++) {
	  status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->u0u[lev][iu],  
			   output->wl.nlambda_s, output->wl.lambda_s, output->u0u[lev][iu], linear);
	  if (status!=0) {
	    fprintf (stderr, " ... error %d interpolating u0u\n", status);
	    return status;
	  }
	}
	
	if ( output->print.phi > 0 ) {
	
	  /* Spline interpolate uu */
	  for (j=0;j<input.rte.nphi;j++) {
	    for (iu=0;iu<input.rte.maxumu;iu++) {
	      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uu[lev][j][iu],  
			       output->wl.nlambda_s, output->wl.lambda_s, output->uu[lev][j][iu], linear);
	      if (status!=0) {
		fprintf (stderr, " ... error %d interpolating uu\n", status);
		return status;
	      }
	    }
	  }
	}
      }
    }
  }
  

  free(output->wl.lambda_h);
  output->wl.nlambda_h = output->wl.nlambda_s;
  output->wl.lambda_h = (float *) calloc (output->wl.nlambda_s, sizeof(float));
  for (iv=0;iv<output->wl.nlambda_h;iv++)
    output->wl.lambda_h[iv] = (float) output->wl.lambda_s[iv];

  return status;
  
}





/**************************************************************/
/* Interpolate the transmittance from the internal grid to    */
/* high resolution grid.                                      */
/**************************************************************/

int interpolate_transmittance (input_struct input, output_struct *output)
{
  int lu=0, iv=0, iu=0, j=0;
  int linear=0, status=0;

  for (lu=0;lu<input.atm.nzout;lu++) {
    
    /**** No need to interpolate if only one wavelength is wanted ****/
    if ( output->wl.nlambda_t == 1 ) {
      iv=0;

      output->flup   [lu][iv] = output->flup_t   [lu][iv];
      output->rfldir [lu][iv] = output->rfldir_t [lu][iv];
      output->rfldn  [lu][iv] = output->rfldn_t  [lu][iv];
      output->uavg   [lu][iv] = output->uavg_t   [lu][iv];
      output->uavgdn [lu][iv] = output->uavgdn_t [lu][iv];
      output->uavgso [lu][iv] = output->uavgso_t [lu][iv];
      output->uavgup [lu][iv] = output->uavgup_t [lu][iv];
      output->sza        [iv] = output->atm.sza_t    [iv];
    } 
    else {
      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->flup_t[lu], 
			output->wl.nlambda_h, output->wl.lambda_h, output->flup[lu], linear);
      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldir_t[lu], 
			output->wl.nlambda_h, output->wl.lambda_h, output->rfldir[lu], linear);
      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldn_t[lu], 
			output->wl.nlambda_h, output->wl.lambda_h, output->rfldn[lu], linear);
      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavg_t[lu], 
			output->wl.nlambda_h, output->wl.lambda_h, output->uavg[lu], linear);
      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgdn_t[lu], 
			output->wl.nlambda_h, output->wl.lambda_h, output->uavgdn[lu], linear);
      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgso_t[lu], 
			output->wl.nlambda_h, output->wl.lambda_h, output->uavgso[lu], linear);
      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgup_t[lu], 
			output->wl.nlambda_h, output->wl.lambda_h, output->uavgup[lu], linear);
      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->atm.sza_t, 
			output->wl.nlambda_h, output->wl.lambda_h, output->sza, linear);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	return status;
      }
    }
    
    for (iu=0;iu<input.rte.maxumu;iu++) {
      
      if (output->wl.nlambda_t == 1) {
	iv = 0;
	output->u0u[lu][iu][iv]      = output->u0u_t[lu][iu][iv];
      } 
      else {
	status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, output->u0u_t[lu][iu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->u0u[lu][iu], linear);

	if (status!=0) {
	  fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	  return status;
	}
      }
    }

    for (j=0;j<input.rte.nphi;j++) {
      for (iu=0;iu<input.rte.maxumu;iu++) {

	if ( output->wl.nlambda_t == 1 ) {
	  iv = 0;
	  output->uu[lu][j][iu][iv] = output->uu_t[lu][j][iu][iv];
	} 
	else {
	  status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, output->uu_t[lu][j][iu], 
			    output->wl.nlambda_h, output->wl.lambda_h, output->uu[lu][j][iu], 
			    linear);
	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	    return status;
	  }
	}
      }
    }
  }

  return 0;
}





/**************************************************************/
/* Mutliply the transmittance with the extraterrestrial       */
/* irradiance.                                                */
/**************************************************************/

int multiply_extraterrestrial (input_struct input, output_struct *output)
{
  int lu=0, iv=0, iu=0, j=0;

  /* multiply with extraterrestrial irradiance */

  if (input.transmittance == 0) {
    for (lu=0;lu<input.atm.nzout;lu++) {
      for (iv=0; iv<output->wl.nlambda_h; iv++) {
	
	/* irradiance and actinic flux */
	output->flup[lu][iv]    *= output->wl.fbeam[iv];
	output->rfldn[lu][iv]   *= output->wl.fbeam[iv];
	output->rfldir[lu][iv]  *= output->wl.fbeam[iv];
	output->uavg[lu][iv]    *= output->wl.fbeam[iv];
	output->uavgdn[lu][iv]  *= output->wl.fbeam[iv];
	output->uavgso[lu][iv]  *= output->wl.fbeam[iv];
	output->uavgup[lu][iv]  *= output->wl.fbeam[iv];
	
	for (iu=0;iu<input.rte.maxumu;iu++) {
	  
	/* azimutally averaged radiance */
	  output->u0u[lu][iu][iv] *= output->wl.fbeam[iv];
	  
	  /* radiance */
	  for (j=0;j<input.rte.nphi;j++)
	    output->uu[lu][j][iu][iv] *= output->wl.fbeam[iv];
	}
      }
    }
  }
  
  return 0;
}



