/************************************************************************/
/*                                                                      */
/* Copyright (C) 1997 Arve Kylling                                      */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*                                                                      */
/************************************************************************/

#include <stdio.h>							    
#include <stdlib.h>
#include <math.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_f/f77-uscore.h"

#define EPSILON 1E-6


/************************************/
/* prototypes of internal functions */
/************************************/

static int shift_cloud (atm_inp_struct *atm_inp, 
			wc_out_struct  *wc_out);
static int compare_levels (float *z1, int n1,
			   float *z2, int n2);




/**************************************************************/
/* Setup water clouds.                                        */
/**************************************************************/

int setup_wcloud (input_struct input, output_struct *output)
{
  int status=0, lev=0, iv=0, rows=0, min_columns=0, max_columns=0;

  float **wc_data=NULL;
  float sum=0, scale_factor=0;

  int nstring = (int) strlen(input.filepath);

  void  F77_FCN (wcloud) (float* lambda_t, int* newsiz, int* nlyr,
			  char* filepath, int* nstring, float* wccon, float* wceffr, 
			  float* tmp_wc_dtau, float* tmp_wc_gg, float* tmp_wc_ssa, float* zd,
			  int *wclyr);


  if ( strlen(input.watercloud_filename) > 0 ) {

    status = ASCII_file2float (input.watercloud_filename, 
			       &rows, &max_columns, &min_columns, 
			       &wc_data);
    if (status!=0) {
      fprintf (stderr, " ... error %d opening file %s\n", status, input.watercloud_filename);
      return status;
    }

    if (max_columns!=min_columns) {
      fprintf (stderr, " !! ATTENTION !! Inconsistent number of columns\n");
      fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
    }  

    if (min_columns<3) {
      fprintf (stderr, " ... error, found less than 3 columns in %s\n",
	       input.watercloud_filename);
      return -1;
    }


    /* allocate memory for 1D cloud structure, output->wc */
    status = calloc_wc_out (&(output->wc), output->wl.nlambda_t, rows); 
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for wc_out_struct\n", status);
      return status;
    }

    for (lev=0; lev<output->wc.wcnlev; lev++) {
      output->wc.wczd  [lev] = wc_data[lev][WCZD];
      output->wc.wccon [lev] = wc_data[lev][WCCON];
      output->wc.wceffr[lev] = wc_data[lev][WCEFFR];
    }

    /* Shift clouds to given altitude */
    if (input.atm.altitude != 0) {
      status = shift_cloud(&input.atm, &output->wc);
      if (status !=0) {
	fprintf (stderr, "error %d shifting clouds\n", status);
	return status;
      }
    }

    /* check for consistency with atmospheric levels */
    status = compare_levels (output->wc.wczd, output->wc.wcnlev,
			     output->atm.zd, output->atm.nlev);
    
    if (status!=0) {
      fprintf (stderr, "Error: z-grids of atmosphere and 1D clouds are different\n");
      return status;
    }
  
    /* newsiz is a flag, telling the Fortran function wcloud() if a */
    /* new liquid water content and/or effective radius are given   */
    /* (newsiz=1) or if the values haven't changed since the last   */
    /* call (newsiz=0).                                             */ 
    output->wc.newsiz = 1;


    for (iv=0; iv<output->wl.nlambda_t; iv++) {

      /* calculate cloud optical properties */
      F77_FCN (wcloud) (&output->wl.lambda_t[iv], &output->wc.newsiz, &output->atm.nlyr,
			input.filepath, &nstring, output->wc.wccon, output->wc.wceffr, 
			output->wc.dtau[iv], output->wc.gg[iv], output->wc.ssa[iv], 
			output->atm.zd, &input.wc.wclayer);
      
      /**** Set the water cloud optical depth to a user specified value ****/
      if (input.wc.wctau >= 0.0) {
	sum = 0.0;
	for (lev=0;lev<output->atm.nlyr;lev++)
	  sum += output->wc.dtau[iv][lev];
	
	if (sum>0)
	  scale_factor = input.wc.wctau/sum;
	else 
	  scale_factor = 0.0;

	for (lev=0; lev<output->atm.nlyr; lev++)
	  output->wc.dtau[iv][lev] *= scale_factor;
      }

      /**** Set the water cloud asymmetry factor to a user specified value ****/
      if ( input.wc.wcgg >= -1.0 )
	for (lev=0;lev<output->atm.nlyr;lev++)
	  output->wc.gg[iv][lev] = input.wc.wcgg;
      
      /**** Set the water cloud single scattering albedo to a user specified value ****/
      if ( input.wc.wcssa >= 0.0 )
	for (lev=0;lev<output->atm.nlyr;lev++)
	  output->wc.ssa[iv][lev] = input.wc.wcssa;
      
    }

    /* free memory */
    status = ASCII_free_float(wc_data, rows);
  }
  else {   
    /* allocate memory, even if no cloud is defined; */
    /* it may be required by other functions.        */

    status = calloc_wc_out (&(output->wc), output->wl.nlambda_t, output->atm.nlev);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for wc_out-struct\n", status);
      return status;
    }
  }
  
  return 0;
}




/**************************************************************/
/* Setup ice clouds.                                          */
/**************************************************************/

int setup_ice (input_struct input, output_struct *output)
{
  int status=0;

  if ((status = ASCII_calloc_float (&output->ice.dtau, output->wl.nlambda_t, output->atm.nlyr)) != 0) 
    return status;
  if ((status = ASCII_calloc_float (&output->ice.gg,   output->wl.nlambda_t, output->atm.nlyr)) != 0) 
    return status;
  if ((status = ASCII_calloc_float (&output->ice.ssa,  output->wl.nlambda_t, output->atm.nlyr)) != 0) 
    return status;

  return 0;
}




/**************************************************************/
/* Shift clouds the same way as the atmosphere                */
/* (see shift_atmosphere).                                    */
/**************************************************************/

static int shift_cloud (atm_inp_struct *atm_inp, 
			wc_out_struct  *wc_out)
{
  int lev=0;

  /* calculate shifted levels (altitude now in km) */
  for (lev=0; lev<wc_out->wcnlev; lev++)
    wc_out->wczd[lev] = wc_out->wczd[lev] - atm_inp->altitude;

  /* check for strange things */
  for (lev=0; lev<wc_out->wcnlev; lev++)
    if (wc_out->wczd[lev]<0 && wc_out->wccon[lev]>0) {
      fprintf (stderr, "Error: the first cloud level is below the surface; you need\n");
      fprintf (stderr, "to change the cloud profile!\n");
      return -1;
    }


  /* determine the first level below 0 */
  lev = wc_out->wcnlev;
  while (wc_out->wczd[--lev]<=0) {
    if (lev<=1) {
      fprintf (stderr, "Sorry, ALTITUDE is higher than the uppermost layer in the\n");
      fprintf (stderr, "water cloud profile.\n");
      return -1;
    }
  }

  /* number of levels is probably smaller now */
  wc_out->wcnlev = lev+2;   /* ??? */

  /* set the last level to 0 */
  wc_out->wczd[wc_out->wcnlev-1] = 0.0; 


  return 0;
}




/*********************************/
/* compare two altitude grids    */
/*********************************/

static int compare_levels (float *z1, int n1,
			   float *z2, int n2)
{
  int kc=0;

  if (n1!=n2) {
    fprintf (stderr, "error, number of layers differing\n");
    fprintf (stderr, "%d in 3D, %d in 1D\n", n1, n2);

    return -1;
  }

  for (kc=0; kc<n1; kc++)
    if (fabs((float) (z1[kc]) - z2[kc]) > EPSILON * z1[kc]) {
      fprintf (stderr, "error, difference at level %d, altitude %g vs. %g\n",	
	       kc, z1[kc], z2[kc]);
      return -1;
    }

  return 0;  /* if o.k. */
}





/***************************************/
/* allocate memory for a wc_out_struct */
/***************************************/

int calloc_wc_out (wc_out_struct *out, int Nlam, int Nlev) 
{
  int iv=0;

  out->dtau = calloc(Nlam, sizeof (float *));
  out->gg   = calloc(Nlam, sizeof (float *));
  out->ssa  = calloc(Nlam, sizeof (float *));
    
  out->wcnlev = Nlev;
  out->wcnlam = Nlam;
  out->newsiz = 1;
  
  out->wczd   = calloc(Nlev, sizeof(float));
  out->wccon  = calloc(Nlev, sizeof(float));
  out->wceffr = calloc(Nlev, sizeof(float));

  for (iv=0; iv<Nlam; iv++) {
    out->dtau[iv] = calloc (Nlev, sizeof(float));
    out->gg  [iv] = calloc (Nlev, sizeof(float));
    out->ssa [iv] = calloc (Nlev, sizeof(float));
  }

  return 0;
}



/************************************************/
/* copy contents of a wc_out_struct to another; */
/* NO memory allocation is done here!!!         */
/************************************************/

int cp_wc_out (wc_out_struct *target, wc_out_struct source) 
{
  int iv=0, lev=0;

  if (target->wcnlev != source.wcnlev) {
    fprintf (stderr, "Error: cp_wc_out(), wcnlev differs between source and target\n");
    fprintf (stderr, "       source: %d,  target: %d\n", source.wcnlev, target->wcnlev);
    return -1;
  }

  if (target->wcnlam != source.wcnlam) {
    fprintf (stderr, "Error: cp_wc_out(), wcnlam differs between source and target\n");
    fprintf (stderr, "       source: %d,  target: %d\n", source.wcnlam, target->wcnlam);
    return -1;
  }

  target->newsiz = source.newsiz;

  /* copy wavelength- and altitude-dependent stuff */
  for (lev=0; lev<source.wcnlev-1; lev++) {
    for (iv=0; iv<source.wcnlam; iv++) {
      target->dtau [iv][lev] = source.dtau [iv][lev];
      target->gg   [iv][lev] = source.gg   [iv][lev];
      target->ssa  [iv][lev] = source.ssa  [iv][lev];
    }
  }
  
  /* copy altitude-dependent stuff */
  for (lev=0; lev<source.wcnlev; lev++) {
    target->wczd   [lev] = source.wczd   [lev];
    target->wccon  [lev] = source.wccon  [lev];
    target->wceffr [lev] = source.wceffr [lev];
  }
  
  return 0;
}
