/*************************************************************************/
/* solve_rte.c                                                           */
/*                                                                       */
/* Various functions helpful for radiative transfer calculations.        */
/*                                                                       */
/* Author:                                                               */
/*   Arve Kylling                                                  @au@  */
/*   NILU,                                                         @ad@  */
/*   Norway.                                                       @ad@  */
/*                                                                       */
/* ----------------------------------------------------------------------*/
/* Copyright (C) 1997 Arve Kylling                                       */
/*                                                                       */
/* This program is free software; you can redistribute it and/or modify  */
/* it under the terms of the GNU General Public License as published by  */
/* the Free Software Foundation; either version 1, or (at your option)   */
/* any later version.                                                    */
/*                                                                       */
/* This program is distributed in the hope that it will be useful,       */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the          */
/* GNU General Public License for more details.                          */
/*                                                                       */
/* To obtain a copy of the GNU General Public License write to the       */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,    */
/* USA.                                                                  */
/*-----------------------------------------------------------------------*/
/*************************************************************************/

#include <math.h>
#include "../src_c/ascii.h"
#include "../src_c/numeric.h"
#include "../src_c/fortran_and_c.h"
#include "../src_c/radiative.h"
#include "../src_c/mc.h"
#include "../src_c/sos.h"
#include "../src_f/f77-uscore.h"
#include "../src_f/solver.h"

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

typedef struct {
  float *dfdt;
  float *flup;
  float *rfldir;
  float *rfldn;
  float *uavgso;
  float *uavgdn;
  float *uavgup;
  float *uavg;
  float *u0u;
  float *uu;
  	 
  float *albmed;
  float *trnmed;
  float *utau;
  float *phi;
} rte_result;




/* prototypes of internal functions */
static int reverse_profiles (input_struct input, output_struct *output);
static rte_result *calloc_rte_result (input_struct input);
static void free_rte_result (rte_result *result);




int solve_rte (input_struct input, output_struct *output)
{
  /* Some disort specific stuff */
  char header[127]; /* Should really be a part of the rte structure */
                    /* but it is not used so who cares */
  float accur = 1.e-3;
  float *beta = NULL;
  float fbeam = 1.0; /* Set to one to normally calculate transmittance, see also fisot */
  float fisot = 0.0;
  float *hl=NULL;
  float *utau=NULL;
  float temis = 0.0;
  float btemp = 0.0;
  float radius= 6370.0; /* Earth's radius in km */
  float *tmp_gg=NULL;
  float ttemp = 0.0;
  float umu0  = 0.0;
  float wvnmhi= 0.0;
  float wvnmlo= 0.0;
  int ierror_d[47] ={0};
  int ierror_s[33] ={0};
  int ierror_t[22] ={0};
  int prndis[7] ={0,0,0,0,0,0,0};
  int prntwo[2] ={0,0};
  int ibcnd=0;
  int imthd=0;
  int lamber=1;
  int newgeo=1;
  int nil=0;
  int onlyfl=1;
  int planck=0; 
  int quiet = 0;
  int spher = 1;
  int usrang = 0;
  int usrtau = 1;

  int status=0;
  int ic=0, ip=0, iu=0, iv=0, j=0, lu=0;
  float *tmp_pmom =NULL, **tmp_u0u_c=NULL, ***tmp_uu_c=NULL;

  rte_result *tmp = calloc_rte_result(input);

  beta = (float *) calloc (output->atm.nlyr+1, sizeof(float));
  hl   = (float *) calloc (input.rte.maxumu+1, sizeof(float));
  utau = (float *) calloc (input.atm.nzout,    sizeof(float));


  /* allocate memory for the output result structures */
  status = setup_result (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for the model output\n", status);
    return status;
  }

  /* Some general initialization before the wavelength loop */
  /* of which some is solver specific                       */
  if (input.rte.fisot > 0) { 
    fisot = 1.0; 
    fbeam = 0.0; 
  }
  else { 
    fisot = 0.0; 
    fbeam = 1.0; 
  }

  if (input.rte.numu > 0 ) { 
    onlyfl = 0; 
    usrang = 1; 
  }

  switch (input.rte.solver) {
  case MONTECARLO:
    break;
  case DISORT:
    for (ip=0; ip<input.rte.nprndis; ip++) 
      prndis[input.rte.prndis[ip]-1] = 1;
    break;
  case SDISORT:
    for (ip=0; ip<input.rte.nprndis; ip++) 
      prndis[input.rte.prndis[ip]-1] = 1;
    break;
  case TWOSTR:
    break;
  case SUCORD:
    break;
  default:
    fprintf (stderr, "Error: RTE solver not yet implemented\n");
    break;
  }    
  
  /* Loop over all wavelenghts, setup optical properties and */
  /* solve rte by the solver the user's choice.              */
  for (iv=0; iv<output->wl.nlambda_t; iv++) {

    umu0 = cos(output->atm.sza_t[iv]*PI/180.);

    /* loop over independent pixels */
    for (ic=0; ic<output->nipa; ic++) {
      
      /* copy pixel ic to 1D cloud data */
      status = cp_wc_out (&(output->wc), output->wcipa[ic]);
      if (status!=0) {
	fprintf (stderr, "Error %d copying output->wcipa[ic] to output->wc\n", status);
	return status;
      }

      /* calculate optical properties from model input */
      status = optical_properties (iv, &input.atm, &output->atm, &output->crs, 
				   output, &input.rte);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by optical_properties()\n", status);
	return status;
      }
      

      /* translate output altitudes (zout) to optical depths (utau) */
      status = set_out (output->dtauc, output->atm.nlyr, input.atm.nzout, 
			utau, output->atm.zd, input.atm.zout );
      if (status!=0) {
	fprintf (stderr, "Error %d returned by set_out()\n", status);
	return status;
      }
      
      /* reverse profiles, if required */
      if (input.rte.reverse) {
	status = reverse_profiles (input, output);
	if (status!=0) {
	  fprintf (stderr, "Error %d reversing profiles\n", status);
	  return status;
	}
      }


      switch (input.rte.solver) {
      case MONTECARLO:
	status = mc (&output->atm.nlyr,
		     output->dtair, output->dto3, output->dtaer,
		     output->omaer, output->fgaer, output->dtcld,
		     output->omcld, output->fgcld, output->dtice,
		     output->omice, output->fgice,
		     output->atm.zmc,
		     &output->atm.albedo_t[iv], &output->atm.sza_t[iv], &output->atm.phi0_t[iv],
		     &input.rte.mcphotons,
		     tmp->rfldir, tmp->rfldn, tmp->flup, 
		     tmp->uavgso, tmp->uavgdn, tmp->uavgup );
	
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by mc()\n", status);
	  return status;
	}
	break;
      case DISORT:
	tmp_pmom = c2fortran_2D_float_ary(output->atm.nlyr, input.rte.nstr+1, output->pmom );
	F77_FCN  (disort)(&output->atm.nlyr, output->dtauc, output->ssalb, tmp_pmom, 
			  &input.rte.nstr, output->atm.temper, &wvnmlo, 
			  &wvnmhi, &usrtau, &input.atm.nzout, utau,
			  &input.rte.nstr, &usrang, &input.rte.numu, input.rte.umu, 
			  &input.rte.nphi, input.rte.phi, &ibcnd, &fbeam, 
			  &umu0, &output->atm.phi0_t[iv], &fisot, &lamber, 
			  &output->atm.albedo_t[iv], hl, &btemp, &ttemp, 
			  &temis, &input.rte.deltam, &imthd, &planck,
			  &onlyfl, &accur, &quiet, ierror_d, 
			  prndis, header, &output->atm.nlyr, 
			  &input.atm.nzout, &input.rte.maxumu, &input.rte.nstr, &input.rte.nstr,
			  &input.rte.maxphi, tmp->rfldir, tmp->rfldn, tmp->flup, 
			  tmp->dfdt, tmp->uavg, 
			  tmp->uu, tmp->u0u, tmp->albmed, tmp->trnmed, 
			  tmp->uavgdn, tmp->uavgso, tmp->uavgup );
	free(tmp_pmom);
	break;
      case SDISORT:
	tmp_pmom = c2fortran_2D_float_ary(output->atm.nlyr, input.rte.nstr+1, output->pmom );
	F77_FCN (sdisort)(&output->atm.nlyr, output->dtauc, output->ssalb, tmp_pmom, 
			  output->atm.temper, &wvnmlo, 
			  &wvnmhi, &usrtau, &input.atm.nzout, utau,
			  &input.rte.nstr, &usrang, &input.rte.numu, input.rte.umu, 
			  &input.rte.nphi, input.rte.phi, &fbeam, beta, &nil,
			  &umu0, &output->atm.phi0_t[iv], &newgeo, output->atm.zd,
			  &spher, &radius, &fisot, &output->atm.albedo_t[iv], 
			  &btemp, &ttemp, &temis, &input.rte.deltam, &planck,
			  &onlyfl, &accur, &quiet, ierror_s, 
			  prndis, header, &output->atm.nlyr, 
			  &input.atm.nzout, &input.rte.maxumu, &input.rte.nstr,
			  &input.rte.maxphi, tmp->rfldir, tmp->rfldn, tmp->flup, 
			  tmp->dfdt, tmp->uavg, 
			  tmp->uu, tmp->u0u, 
			  tmp->uavgdn, tmp->uavgso, tmp->uavgup );
	free(tmp_pmom);
	break;
      case TWOSTR:
	tmp_gg = (float *) calloc (output->atm.nlyr, sizeof(float));
	for (lu=0; lu<output->atm.nlyr; lu++) 
	  tmp_gg[lu]=output->pmom[lu][1];

	F77_FCN  (twostr)( &output->atm.albedo_t[iv], &btemp, &input.rte.deltam,
			   output->dtauc, &fbeam, &fisot, tmp_gg, header, 
			   ierror_t, &output->atm.nlyr, &input.atm.nzout, &newgeo, 
			   &output->atm.nlyr, &planck, &input.atm.nzout,
			   prntwo, &quiet, &radius, &spher, output->ssalb, &temis,  
			   output->atm.temper, &ttemp, &umu0, &usrtau, utau, &wvnmlo,
			   &wvnmhi, output->atm.zd, tmp->dfdt, tmp->flup, 
			   tmp->rfldir, tmp->rfldn, tmp->uavg);
	free(tmp_gg);
	break;
      case SUCORD:
	status = sos (output->atm.nlyr, newgeo, input.rte.nstr, input.rte.nscat,
		      radius, output->atm.zd, output->ssalb, output->pmom,
		      output->dtauc, 
		      output->atm.sza_t[iv], input.atm.nzout, 
		      utau, 
		      tmp->rfldir, tmp->rfldn, tmp->flup, 
		      tmp->uavgso, tmp->uavgdn, tmp->uavgup );
	
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by sos()\n", status);
	  return status;
	}
	break;
      default:
	fprintf (stderr, "Error: RTE solver not yet implemented\n");
	return -1;
      }
      
      
      tmp_u0u_c = fortran2c_2D_float_ary(input.atm.nzout, input.rte.maxumu, tmp->u0u);
      tmp_uu_c  = fortran2c_3D_float_ary(input.rte.maxphi, input.atm.nzout, 
					 input.rte.maxumu, tmp->uu);



      /* print the results for pixel ic */
      /*
      fprintf (stderr, "%4i %8.6f %7.3f %13.6e %13.6e %13.6e %13.6e %13.6e %13.6e\n",
	       ic, output->ipaweight[ic], output->wl.lambda_t[iv],
	       tmp->rfldir[0], tmp->rfldn[0], tmp->flup[0],
	       tmp->uavgso[0], tmp->uavgdn[0], tmp->uavgup[0]);
      */

      /* add result for the current pixel */
      for(lu=0; lu<input.atm.nzout; lu++) {
	output->rfldir_t [lu][iv] += output->ipaweight[ic]*tmp->rfldir [lu];
	output->rfldn_t  [lu][iv] += output->ipaweight[ic]*tmp->rfldn  [lu];
	output->flup_t   [lu][iv] += output->ipaweight[ic]*tmp->flup   [lu];
	output->uavg_t   [lu][iv] += output->ipaweight[ic]*tmp->uavg   [lu];
	output->uavgdn_t [lu][iv] += output->ipaweight[ic]*tmp->uavgdn [lu];
	output->uavgso_t [lu][iv] += output->ipaweight[ic]*tmp->uavgso [lu];
	output->uavgup_t [lu][iv] += output->ipaweight[ic]*tmp->uavgup [lu];
	
	for (iu=0;iu<input.rte.numu;iu++) {
	  output->u0u_t[lu][iu][iv] += output->ipaweight[ic]*tmp_u0u_c[lu][iu];
	  
	  for (j=0;j<input.rte.nphi;j++)
	    output->uu_t[lu][j][iu][iv] += output->ipaweight[ic]*tmp_uu_c[j][lu][iu];
	}
      }


      for(lu=0; lu<input.atm.nzout; lu++) {
	free (tmp_u0u_c[lu]);
	
	/* ??? is input.rte.maxphi the right stop ??? */
	for (j=0; j<input.rte.maxphi; j++)
	  free(tmp_uu_c[j][lu]);
      }      
      
      free(tmp_u0u_c);
      free(tmp_uu_c);
    }
  }


  /* free temporary memory */
  free_rte_result (tmp);



  return 0;
}





int setup_result (input_struct input, output_struct *output)
{
  int status=0;

  if ((status = ASCII_calloc_float (&output->flup_t,   input.atm.nzout, output->wl.nlambda_t)) != 0)
    return status;

  if ((status = ASCII_calloc_float (&output->rfldir_t, input.atm.nzout, output->wl.nlambda_t)) != 0)
    return status;

  if ((status = ASCII_calloc_float (&output->rfldn_t,  input.atm.nzout, output->wl.nlambda_t)) !=0 )
    return status;
  
  if ((status = ASCII_calloc_float (&output->uavg_t,   input.atm.nzout, output->wl.nlambda_t)) !=0 )
    return status;

  if ((status = ASCII_calloc_float (&output->uavgdn_t, input.atm.nzout, output->wl.nlambda_t)) !=0 )
    return status;

  if ((status = ASCII_calloc_float (&output->uavgso_t, input.atm.nzout, output->wl.nlambda_t)) !=0 )
    return status;

  if ((status = ASCII_calloc_float (&output->uavgup_t, input.atm.nzout, output->wl.nlambda_t)) !=0 )
    return status;

  if ((status = ASCII_calloc_float_3D (&output->u0u_t, input.atm.nzout, input.rte.maxumu,
				       output->wl.nlambda_t)) != 0)
    return status;
  
  if ((status = ASCII_calloc_float_4D (&output->uu_t,  input.atm.nzout, input.rte.maxphi, 
				       input.rte.maxumu, output->wl.nlambda_t)) !=0)
    return status;
  


  
  if ((status = ASCII_calloc_float (&output->flup,   input.atm.nzout, output->wl.nlambda_h)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->rfldn,  input.atm.nzout, output->wl.nlambda_h)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->rfldir, input.atm.nzout, output->wl.nlambda_h)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavg,   input.atm.nzout, output->wl.nlambda_h)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavgdn, input.atm.nzout, output->wl.nlambda_h)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavgso, input.atm.nzout, output->wl.nlambda_h)) != 0) 
    return status;

  if ((status = ASCII_calloc_float (&output->uavgup, input.atm.nzout, output->wl.nlambda_h)) != 0) 
    return status;
  
  if ((status = ASCII_calloc_float_3D (&output->u0u, input.atm.nzout, input.rte.maxumu, 
				       output->wl.nlambda_h)) != 0) 
    return status;

  if ((status = ASCII_calloc_float_4D (&output->uu, input.atm.nzout, input.rte.maxphi, 
				       input.rte.maxumu, output->wl.nlambda_h)) != 0) 
    return status;

  if ((output->sza = (float *) calloc (output->wl.nlambda_h, sizeof(float))) == NULL)
    return -1;
  
  return 0;
}

static int reverse_profiles (input_struct input, output_struct *output)
{
  int lu=0, iu=0;
  double tmp=0;

  for (lu=0;lu<output->atm.nlyr/2;lu++) {
    
    /* reverse profile of optical depth dtauc */
    tmp               = output->dtauc[output->atm.nlyr-lu-1];
    output->dtauc[output->atm.nlyr-lu-1] = output->dtauc[lu];
    output->dtauc[lu] = tmp;
    
    /* reverse profile of single scattering albedo ssalb */
    tmp               = output->ssalb[output->atm.nlyr-lu-1];
    output->ssalb[output->atm.nlyr-lu-1] = output->ssalb[lu];
    output->ssalb[lu] = tmp;
    
    /* reverse profile of phase function pmom */
    for (iu=0;iu<=input.rte.nstr;iu++) {
      tmp                  =  output->pmom[output->atm.nlyr-lu-1][iu];
      output->pmom[output->atm.nlyr-lu-1][iu] = output->pmom[lu][iu];
      output->pmom[lu][iu] = tmp;
    }
  }
  return 0;
}

/* allocate temporary memory for the RTE solvers */
static rte_result *calloc_rte_result (input_struct input)
{
  rte_result *result = calloc(1, sizeof(rte_result));
  
  result->albmed = (float *) calloc (input.rte.maxumu, sizeof(float));
  result->trnmed = (float *) calloc (input.rte.maxumu, sizeof(float));
  result->phi    = (float *) calloc (input.rte.maxphi, sizeof(float));
  result->utau   = (float *) calloc (input.atm.nzout,  sizeof(float));
	     
  result->dfdt   = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->flup   = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->rfldir = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->rfldn  = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->uavg   = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->uavgdn = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->uavgso = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->uavgup = (float *) calloc (input.atm.nzout,  sizeof(float));
  result->u0u    = (float *) calloc (input.atm.nzout*input.rte.maxumu, sizeof(float));
  result->uu     = (float *) calloc (input.atm.nzout*input.rte.maxumu*input.rte.maxphi, 
				     sizeof(float));
  return result;
}




/* free temporary memory for the RTE solvers */
static void free_rte_result (rte_result *result)
{
  
  free(result->albmed);
  free(result->trnmed);
  free(result->phi);
  free(result->utau);
	     
  free(result->dfdt);
  free(result->flup);
  free(result->rfldir);
  free(result->rfldn);
  free(result->uavg);
  free(result->uavgdn);
  free(result->uavgso);
  free(result->uavgup);
  free(result->u0u);
  free(result->uu);

  free(result);
}
